# Copyright 2015 Mirantis Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import fixture as config
from oslotest import base as test_base

from oslo_service import _options


class ServiceBaseTestCase(test_base.BaseTestCase):

    def setUp(self):
        super(ServiceBaseTestCase, self).setUp()
        self.conf_fixture = self.useFixture(config.Config())
        self.conf_fixture.register_opts(_options.eventlet_backdoor_opts)
        self.conf_fixture.register_opts(_options.service_opts)
        self.conf_fixture.register_opts(_options.ssl_opts)
        self.conf_fixture.register_opts(_options.periodic_opts)

        self.conf = self.conf_fixture.conf
        self.config = self.conf_fixture.config
